"""Data conversion unit tests.

"""

# Native imports

# Third-party imports
import gsw
import numpy as np
import pandas as pd
import pytest

# Sea-Bird imports

# Internal imports
# import seabirdscientific.cal_coefficients as cc
import seabirdscientific.conversion as dc
import seabirdscientific.instrument_data as id
import seabirdscientific.contour as c

import example_coefficients as ec

# DATASETS = [
#     # cal
#     # cnv
#     # hex
#     # cond_label
#     # use_MV_R
#     (
#         cc.SN6130,
#         id.cnv_to_instrument_data("./tests/resources/test-data/19plus_V2_CTD-processing_example.cnv"),
#         id.read_hex_file("./documentation/example_data/19plus_V2_CTD-processing_example.hex",
#             id.InstrumentType.SBE19Plus,
#             [
#                 id.Sensors.Temperature,
#                 id.Sensors.Conductivity,
#                 id.Sensors.Pressure,
#                 id.Sensors.ExtVolt0,
#                 id.Sensors.ExtVolt1,
#                 id.Sensors.ExtVolt2,
#                 id.Sensors.ExtVolt4,
#             ],
#         ),
#         "c0S/m",
#         True,
#     ),
#     # removing until we get a smaller version of this source (NSI-3079)
#     # (
#     #     cc.SN03706385,
#     #     id.cnv_to_instrument_data("./tests/resources/test-data/SBE37SM.cnv"),
#     #     id.read_hex_file(
#     #         "./tests/resources/test-data/SBE37SM.hex",
#     #         id.InstrumentType.SBE37SM,
#     #         [
#     #             id.Sensors.Temperature,
#     #             id.Sensors.Conductivity,
#     #             id.Sensors.Pressure,
#     #         ],
#     #     ),
#     #     "cond0S/m",
#     #     False,
#     # ),
#     (
#         cc.SN03716125,
#         id.cnv_to_instrument_data("./tests/resources/test-data/SBE37SM-RS232_03716125_2017_11_16.cnv"
#         ),
#         id.read_hex_file("./tests/resources/test-data/SBE37SM-RS232_03716125_2017_11_16.hex",
#             id.InstrumentType.SBE37SM,
#             [
#                 id.Sensors.Temperature,
#                 id.Sensors.Conductivity,
#                 id.Sensors.Pressure,
#             ],
#         ),
#         "cond0S/m",
#         False,
#     ),
# ]


class TestConvertTemperature:
    # test temperature raw values
    test_temp_vals = np.array([322798, 322808, 322827, 322838])

    def test_convert_temperature_90C(self):
        expected = [20.4459, 20.4451, 20.4436, 20.4427]
        result = dc.convert_temperature(
            self.test_temp_vals,
            ec.temperature_coefs_sn6130,
            "ITS90",
            "C",
            True,
        )
        assert np.allclose(expected, result, rtol=0, atol=1e-4)

    def test_convert_temperature_90F(self):
        expected = [68.8027, 68.8012, 68.7984, 68.7968]
        result = dc.convert_temperature(
            self.test_temp_vals,
            ec.temperature_coefs_sn6130,
            "ITS90",
            "F",
            True,
        )
        assert np.allclose(expected, result, rtol=0, atol=1e-4)

    def test_convert_temperature_68C(self):
        expected = [20.4508, 20.4500, 20.4485, 20.4476]
        result = dc.convert_temperature(
            self.test_temp_vals,
            ec.temperature_coefs_sn6130,
            "IPTS68",
            "C",
            True,
        )
        assert np.allclose(expected, result, rtol=0, atol=1e-4)

    def test_convert_temperature_68F(self):
        expected = [68.8115, 68.8100, 68.8073, 68.8057]
        result = dc.convert_temperature(
            self.test_temp_vals,
            ec.temperature_coefs_sn6130,
            "IPTS68",
            "F",
            True,
        )
        assert np.allclose(expected, result, rtol=0, atol=1e-4)


class TestConvertPressure:
    # test pressure raw values
    test_pressure_vals = np.array([533539, 533538, 533540, 533537])

    # test pressure temperature compensation raw values
    test_compensation_vals = np.array([20625, 20626, 20626, 20626]) / 13107

    def test_convert_pressure_psia(self):
        expected = [-0.153, -0.154, -0.151, -0.156]
        result = dc.convert_pressure(
            self.test_pressure_vals,
            self.test_compensation_vals,
            ec.pressure_coefs_sn6130,
            "psia",
        )
        assert np.allclose(expected, result, rtol=0, atol=1e-3)

    def test_convert_pressure_dbar(self):
        expected = [-0.105, -0.106, -0.104, -0.107]
        result = dc.convert_pressure(
            self.test_pressure_vals,
            self.test_compensation_vals,
            ec.pressure_coefs_sn6130,
            "dbar",
        )
        assert np.allclose(expected, result, rtol=0, atol=1e-3)


class TestConductivity19plus:
    cnv_path = "./documentation/example_data/19plus_V2_CTD-processing_example.hex"
    expected_data = id.cnv_to_instrument_data(cnv_path)

    hex_path = "./documentation/example_data/19plus_V2_CTD-processing_example.hex"
    raw = id.read_hex_file(
        hex_path,
        id.InstrumentType.SBE19Plus,
        [
            id.Sensors.Temperature,
            id.Sensors.Conductivity,
            id.Sensors.Pressure,
            id.Sensors.ExtVolt0,
            id.Sensors.ExtVolt1,
            id.Sensors.ExtVolt2,
            id.Sensors.ExtVolt4,
        ],
    )

    # test_cond_data = np.array([675415, 675404, 675405, 675398, 675391, 675379])

    # test_temp_data = np.array([20.4459, 20.4451, 20.4436, 20.4427, 20.4413, 20.4401])
    temperature = dc.convert_temperature(
        raw["temperature"][0:6].values,
        ec.temperature_coefs_sn6130,
        "IPTS68",
        "C",
        True,
    )

    # test_press_data = np.array([-0.105, -0.106, -0.104, -0.107, -0.105, -0.104])
    pressure = dc.convert_pressure(
        raw["pressure"][0:6].values,
        raw["temperature compensation"][0:6].values,
        ec.pressure_coefs_sn6130,
        "psia",
    )

    def test_convert_conductivity(self):
        expected = [0.008453, 0.008420, 0.008423, 0.008402, 0.008380, 0.008344]
        result = dc.convert_conductivity(
            self.raw["conductivity"][0:6].values,
            self.temperature,
            self.pressure,
            ec.conductivity_coefs_sn6130,
        )
        assert np.allclose(expected, result, rtol=0, atol=1e-6)


class TestConductivity37SM:
    cnv_path = "./tests/resources/test-data/SBE37SM-RS232_03716125_2017_11_16.cnv"
    expected_data = id.cnv_to_instrument_data(cnv_path)

    hex_path = "./tests/resources/test-data/SBE37SM-RS232_03716125_2017_11_16.hex"
    raw = id.read_hex_file(
        hex_path,
        id.InstrumentType.SBE37SM,
        [
            id.Sensors.Temperature,
            id.Sensors.Conductivity,
            id.Sensors.Pressure,
        ],
    )

    temperature = dc.convert_temperature(
        raw["temperature"][0:6],
        ec.temperature_coefs_sn6130,
        "IPTS68",
        "C",
        True,
    )

    # test_press_data = np.array([-0.105, -0.106, -0.104, -0.107, -0.105, -0.104])
    pressure = dc.convert_pressure(
        raw["pressure"][0:6],
        raw["temperature compensation"][0:6],
        ec.pressure_coefs_sn16125,
        "psia",
    )

    def test_convert_conductivity(self):
        expected = [2.711842, 2.715786, 2.715857, 2.715846, 2.715846, 2.715857]
        result = dc.convert_conductivity(
            self.raw["conductivity"][0:6],
            self.temperature,
            self.pressure,
            ec.conductivity_coefs_sn16125,
        )
        assert np.allclose(expected, result, rtol=0, atol=1e-4)


class TestDeriveDensity:
    data_path = "./tests/resources/test-data/SBE37SM-derived.asc"
    data = pd.read_csv(data_path)

    @pytest.mark.parametrize(
        "reference_pressure, expected_column",
        [
            (0.0, "gsw_sigma0A0"),
            (1000.0, "gsw_sigma1A0"),
            (2000.0, "gsw_sigma2A0"),
            (3000.0, "gsw_sigma3A0"),
            (4000.0, "gsw_sigma4A0"),
        ],
    )
    def test_derive_potential_density_from_t_s_p_pass(self, reference_pressure, expected_column):
        temperature = self.data["t090C"].values
        salinity = self.data["sal00"].values
        pressure = self.data["prM"].values
        expected = self.data[expected_column].values
        result = dc.potential_density_from_t_s_p(
            temperature,
            salinity,
            pressure,
            reference_pressure=reference_pressure,
        )
        # TODO: improve passing condition for all instances of allclose
        assert np.allclose(result, expected, atol=0.1)

    def test_derive_density_from_t_s_p_pass(self):
        temperature = self.data["t090C"].values
        salinity = self.data["sal00"].values
        pressure = self.data["prM"].values
        expected = self.data["gsw_densityA0"].values
        result = dc.density_from_t_s_p(temperature, salinity, pressure)

        assert np.allclose(result, expected, atol=0.1)

    @pytest.mark.parametrize(
        "reference_pressure, expected_column",
        [
            (0.0, "gsw_sigma0A0"),
            (1000.0, "gsw_sigma1A0"),
            (2000.0, "gsw_sigma2A0"),
            (3000.0, "gsw_sigma3A0"),
            (4000.0, "gsw_sigma4A0"),
        ],
    )
    def test_derive_potential_density_from_t_c_p_pass(self, reference_pressure, expected_column):
        temperature = self.data["t090C"].values
        conductivity = self.data["c0S/m"].values * 10.0
        pressure = self.data["prM"].values
        expected = self.data[expected_column].values
        result = dc.potential_density_from_t_c_p(
            temperature,
            conductivity,
            pressure,
            reference_pressure=reference_pressure,
        )
        assert np.allclose(result, expected, atol=0.1)

    def test_derive_density_from_t_c_p_pass(self):
        temperature = self.data["t090C"].values
        conductivity = self.data["c0S/m"].values * 10.0
        pressure = self.data["prM"].values
        expected = self.data["gsw_densityA0"].values
        result = dc.density_from_t_c_p(temperature, conductivity, pressure)

        assert np.allclose(result, expected, atol=0.1)


class TestDepthFromPressure:
    data_path = "./tests/resources/test-data/SBE37SM.asc"
    data = pd.read_csv(data_path).loc[6:10]

    @pytest.mark.parametrize(
        "pressure, pressure_units, depth, depth_units",
        [
            ("prdM", "dbar", "depSM", "m"),
            ("prdM", "dbar", "depSF", "ft"),
            ("prdE", "psi", "depSM", "m"),
            ("prdE", "psi", "depSF", "ft"),
        ],
    )
    def test_depth_from_pressure_pass(self, pressure, pressure_units, depth, depth_units):
        expected_depth = self.data[depth].values
        pressure = self.data[pressure].values
        result_depth = dc.depth_from_pressure(pressure, 0, depth_units, pressure_units)
        assert np.allclose(expected_depth, result_depth, atol=0.002)


# class TestSalinityFromTCP:
#     @pytest.mark.parametrize("cal, cnv, hex, cond_label, use_MV_R", DATASETS)
#     def test_salinity_from_tcp(self, cal, cnv, hex, cond_label, use_MV_R):
#         expected_salinity = cnv.measurements["sal00"].values[800:810]
#         temperature = cnv.measurements["tv290C"].values[800:810]
#         pressure = cnv.measurements["prdM"].values[800:810]
#         conductivity = cnv.measurements[cond_label].values[800:810] * 10
#         result_salinity = gsw.SP_from_C(conductivity, temperature, pressure)
#         assert np.allclose(expected_salinity, result_salinity, rtol=0, atol=1e-4)

#     @pytest.mark.parametrize("cal, cnv, hex, cond_label, use_MV_R", DATASETS)
#     def test_salinity_from_tcp_raw(self, cal, cnv, hex, cond_label, use_MV_R):
#         """Converts data from raw hex values.

#         The unused data loaded from cnv are not needed for the test,
#         but are useful for comparing during a debug session.

#         Args:
#             parameterized from the DATASET const at the top of the file
#         """

#         salinity_expected = cnv.measurements["sal00"].values[800:810]
#         temperature = cnv.measurements["tv290C"].values[800:810]
#         temperature_raw = hex.temperature.values[800:810]
#         temperature_conv = dc.convert_temperature_array(
#             temperature_raw, cal.A0, cal.A1, cal.A2, cal.A3, True, True, use_MV_R
#         )

#         pressure = cnv.measurements["prdM"].values[800:810]
#         pressure_raw = hex.pressure.values[800:810]
#         other_raw = hex.iloc[:, 3].values[800:810]  # either salinity or pressure temperature compensation
#         pressure_conv = dc.convert_pressure_array(
#             pressure_raw,
#             other_raw,
#             True,
#             cal.PA0,
#             cal.PA1,
#             cal.PA2,
#             cal.PTEMPA0,
#             cal.PTEMPA1,
#             cal.PTEMPA2,
#             cal.PTCA0,
#             cal.PTCA1,
#             cal.PTCA2,
#             cal.PTCB0,
#             cal.PTCB1,
#             cal.PTCB2,
#         )

#         conductivity = cnv.measurements[cond_label].values[800:810] * 10
#         conductivity_raw = hex.conductivity.values[800:810]
#         conductivity_conv = (
#             dc.convert_conductivity_array(
#                 conductivity_raw,
#                 temperature_conv,
#                 pressure_conv,
#                 cal.G,
#                 cal.H,
#                 cal.I,
#                 cal.J,
#                 cal.CPCOR,
#                 cal.CTCOR,
#                 cal.WBOTC,
#             )
#             * 10
#         )
#         # can't un-round cnv data, so instead validate that converted hex data rounds to same value
#         salinity_result = np.round(gsw.SP_from_C(conductivity_conv, temperature_conv, pressure_conv), 4)
#         assert np.allclose(salinity_expected, salinity_result, rtol=0, atol=1e-4)


class TestConvertOxygen:
    def test_convert_sbe63_oxygen(self):
        raw_oxygen = np.array([31.06, 31.66, 32.59, 33.92, 34.82, 35.44])
        pressure = np.array([0, 0, 0, 0, 0, 0])
        raw_temperature = np.array([0.6, 0.5, 0.4, 0.35, 0.3, 0.25])
        salinity = np.array([0, 0, 0, 0, 0, 0])
        expected = np.array([0.93, 0.688, 0.459, 0.304, 0.206, 0.137])

        result = dc.convert_sbe63_oxygen(
            raw_oxygen,
            raw_temperature,
            pressure,
            salinity,
            ec.oxygen_63_coefs_sn2568,
            ec.thermistor_63_coefs_sn2568,
        )
        assert np.allclose([expected], [result], rtol=0, atol=1e-2)

    def test_convert_sbe43_oxygen(self):
        # From O3287.pdf in the shared calibration folder
        raw_oxygen = np.array([0.725, 0.756, 0.803, 0.874, 0.925, 0.96, 1.332, 1.435, 1.595, 1.81])
        pressure = np.array([0, 0, 0, 0, 0, 0, 0, 0, 0, 0])
        temperature = np.array([2, 6, 12, 20, 26, 30, 2, 6, 12, 20])
        salinity = np.array([0, 0, 0, 0, 0, 0, 0, 0, 0, 0])
        expected = np.array([1.11, 1.12, 1.13, 1.16, 1.18, 1.18, 3.9, 3.9, 3.92, 3.95])

        result = dc._convert_sbe43_oxygen(
            raw_oxygen,
            temperature,
            pressure,
            salinity,
            ec.oxygen_43_coefs_sn3287,
            0,
        )
        assert np.allclose([expected], [result], rtol=0, atol=1e-2)

    def test_convert_sbe43_oxygen_from_hex(self):
        # From SBE19plus_01906398_2019_07_15_0033.hex
        raw_oxygen = np.array(
            [
                2.5575,
                2.5586,
                2.5606,
                2.5627,
                2.5638,
                2.5637,
                2.5635,
                2.5629,
                2.5621,
                2.5618,
            ]
        )
        pressure = np.array(
            [-0.012, -0.012, -0.012, -0.012, -0.012, -0.012, -0.012, -0.011, 0.107, 0.351]
        )
        temperature = np.array(
            [
                25.3427,
                25.3408,
                25.3387,
                25.3363,
                25.3341,
                25.3326,
                25.3316,
                25.3302,
                25.3377,
                25.5433,
            ]
        )
        salinity = np.array(
            [0.4373, 0.5592, 0.5865, 0.5095, 0.4621, 0.4119, 0.3936, 0.3463, 4.9297, 6.5098]
        )
        expected = np.array(
            [4.4728, 4.4722, 4.4762, 4.4828, 4.4867, 4.4879, 4.488, 4.488, 4.3707, 4.3148]
        )
        for index in range(len(expected)):
            result = dc._convert_sbe43_oxygen(
                raw_oxygen[index],
                temperature[index],
                pressure[index],
                salinity[index],
                ec.oxygen_43_coefs_sn1686,
                0,
            )
            assert np.allclose([expected[index]], [result], rtol=0, atol=1e-3)

    def test_convert_sbe43_oxygen_from_hex_with_hysteresis(self):
        # TODO: This test is failing. Fix as part of NSI-3061
        # From SBE19plus_01906398_2019_07_15_0033.hex
        # TODO: hysteresis correction only has a real impact on deep data, will need some to better validate this
        raw_oxygen = np.array(
            [
                2.5575,
                2.5586,
                2.5606,
                2.5627,
                2.5638,
                2.5637,
                2.5635,
                2.5629,
                2.5621,
                2.5618,
            ]
        )
        pressure = np.array(
            [-0.012, -0.012, -0.012, -0.012, -0.012, -0.012, -0.012, -0.011, 0.107, 0.351]
        )
        temperature = np.array(
            [
                25.3427,
                25.3408,
                25.3387,
                25.3363,
                25.3341,
                25.3326,
                25.3316,
                25.3302,
                25.3377,
                25.5433,
            ]
        )
        salinity = np.array(
            [0.4373, 0.5592, 0.5865, 0.5095, 0.4621, 0.4119, 0.3936, 0.3463, 4.9297, 6.5098]
        )
        expected = np.array(
            [4.4728, 4.4722, 4.4762, 4.4828, 4.4867, 4.4879, 4.488, 4.488, 4.3707, 4.3148]
        )
        result = dc.convert_sbe43_oxygen(
            raw_oxygen,
            temperature,
            pressure,
            salinity,
            ec.oxygen_43_coefs_sn1686,
            False,
            True,
            1,
            0.25,
        )

        assert np.allclose(result, expected, rtol=0, atol=1e-3)

    def test_convert_sbe43_oxygen_from_hex_with_tau_correction(self):
        # From SBE19plus_01906398_2019_07_15_0033.hex
        raw_oxygen = np.asarray(
            [2.5575, 2.5586, 2.5606, 2.5627, 2.5638, 2.5637, 2.5635, 2.5629, 2.5621, 2.5618]
        )
        pressure = np.asarray(
            [-0.012, -0.012, -0.012, -0.012, -0.012, -0.012, -0.012, -0.011, 0.107, 0.351]
        )
        temperature = np.asarray(
            [
                25.3427,
                25.3408,
                25.3387,
                25.3363,
                25.3341,
                25.3326,
                25.3316,
                25.3302,
                25.3377,
                25.5433,
            ]
        )
        salinity = np.asarray(
            [0.4373, 0.5592, 0.5865, 0.5095, 0.4621, 0.4119, 0.3936, 0.3463, 4.9297, 6.5098]
        )
        expected = [4.4729, 4.4723, 4.4884, 4.4927, 4.4916, 4.4879, 4.4849, 4.4841, 4.3707, 4.3148]

        result = dc.convert_sbe43_oxygen(
            raw_oxygen,
            temperature,
            pressure,
            salinity,
            ec.oxygen_43_coefs_sn1686,
            True,
            False,
            1,
            0.25,
        )

        assert np.allclose(result, expected, rtol=0, atol=1e-4)

    def test_convert_to_mg_per_l(self):
        oxMlPerL = np.array(
            [4.4728, 4.4722, 4.4762, 4.4828, 4.4867, 4.4879, 4.488, 4.488, 4.3707, 4.3148]
        )
        expected = [6.3921, 6.3913, 6.3969, 6.4064, 6.4119, 6.4137, 6.4138, 6.4138, 6.2461, 6.1663]
        result = dc.convert_oxygen_to_mg_per_l(oxMlPerL)
        for index in range(len(expected)):
            assert np.allclose([expected[index]], [result[index]], rtol=0, atol=1e-3)

    def test_convert_to_umol_per_kg(self):
        oxMlPerL = np.array(
            [4.4728, 4.4722, 4.4762, 4.4828, 4.4867, 4.4879, 4.488, 4.488, 4.3707, 4.3148]
        )
        expected = [
            200.3,
            200.254,
            200.427,
            200.735,
            200.916,
            200.979,
            200.984,
            200.991,
            195.064,
            192.356,
        ]
        potentialDensity = np.array(
            [
                -2.7113,
                -2.6188,
                -2.5977,
                -2.6552,
                -2.6903,
                -2.7279,
                -2.7414,
                -2.7768,
                0.6655,
                1.7939,
            ]
        )
        result = dc.convert_oxygen_to_umol_per_kg(oxMlPerL, potentialDensity)
        for index in range(len(expected)):
            assert np.allclose([expected[index]], [result[index]], rtol=0, atol=1e-2)


class TestConvertChlorophylla:
    def test_convert_eco(self):
        rawAnalog = np.array(
            [
                0.0949,
                0.0948,
                0.0960,
                0.0961,
                0.0962,
                0.0959,
                0.1013,
                0.1012,
                0.1015,
                0.1012,
                0.1003,
                0.0999,
                0.0999,
                0.0996,
            ]
        )
        expected = np.array(
            [
                0.2691,
                0.2683,
                0.2798,
                0.2813,
                0.2821,
                0.279,
                0.3332,
                0.3317,
                0.3355,
                0.3317,
                0.3233,
                0.3187,
                0.3195,
                0.3157,
            ]
        )
        for index in range(len(expected)):
            result = dc.convert_eco(rawAnalog[index], ec.chlorophyll_a_coefs_sn6130)
            assert np.allclose([expected[index]], [result], rtol=0, atol=1e-2)


class TestConvertpH:
    def test_convert_sbe18_ph(self):
        rawVolts = np.array(
            [
                2.9507,
                2.9518,
                2.9522,
                2.9517,
                2.9521,
                2.9519,
                2.9517,
                2.9526,
                2.9525,
                2.9537,
                2.953,
                2.9527,
            ]
        )
        temperatureC = np.array(
            [
                26.6927,
                26.6994,
                26.6624,
                26.6122,
                26.5808,
                26.5621,
                26.5503,
                26.5505,
                26.569,
                26.5954,
                26.6182,
                26.6359,
            ]
        )
        expected = np.array(
            [
                8.587,
                8.591,
                8.593,
                8.591,
                8.593,
                8.592,
                8.592,
                8.595,
                8.594,
                8.599,
                8.596,
                8.595,
            ]
        )
        for index in range(len(expected)):
            result = dc.convert_sbe18_ph(rawVolts[index], temperatureC[index], ec.ph_coefs_sn0762)
            assert np.allclose([expected[index]], [result], rtol=0, atol=1e-3)


class TestPARlogarithmic:
    def test_convert_par_logarithmic(self):
        rawVolts = np.array(
            [
                1.176241702,
                1.175249866,
                1.174944686,
                1.176394293,
                0.089951934,
                0.090562294,
                0.090638590,
                0.091172655,
                0.090485999,
                0.090943770,
                0.090409704,
                0.091096360,
                0.093537804,
            ]
        )
        expected = np.array(
            [
                0.81605,
                0.81394,
                0.81330,
                0.81637,
                0.04817,
                0.04824,
                0.04825,
                0.04832,
                0.04823,
                0.04829,
                0.04822,
                0.04831,
                0.04862,
            ]
        )
        for index in range(len(expected)):
            result = dc.convert_par_logarithmic(
                rawVolts[index],
                ec.par_coefs_sn0411,
            )
            assert np.allclose([expected[index]], [result], rtol=0, atol=1e-3)


# class TestContourFromTSP:
#     # Note: this class doesn't actually test anything and is only for debug
#     data_path = './tests/resources/test-data/SBE37SM-RS232_03711722_2015_11_18_subset_derived.asc'
#     data = pd.read_csv(data_path)

#     def test_contour_from_t_s_p_pass(self):
#         temperature = self.data['t090C'].values
#         salinity = self.data['sal00'].values
#         pressure = self.data['prM'].values
#         contour_data = c.contour_from_t_s_p(
#             temperature, salinity, pressure
#         )
#         assert(True)


class TestSeaFETPH:
    internal_ph_counts = np.array([5105334, 5105384, 5105350, 5105505, 5105347])
    external_ph_counts = np.array([5139133, 5139214, 5139120, 5139294, 5139183])
    expected_internal_ph_volts = np.array([-0.978492, -0.978477, -0.978487, -0.978441, -0.978488])
    expected_external_ph_volts = np.array([-0.968419, -0.968395, -0.968423, -0.968371, -0.968404])
    expected_internal_ph = np.array([6.8897, 6.89, 6.8898, 6.8906, 6.8898])
    expected_external_ph = np.array([6.2466, 6.247, 6.2466, 6.2474, 6.2469])
    ph_temperature = np.array([22.751, 22.7514, 22.752, 22.752, 22.752])

    def test_convert_ph_voltage_counts(self):
        internal_ph_volts = dc.convert_ph_voltage_counts(self.internal_ph_counts)
        assert np.allclose(internal_ph_volts, self.expected_internal_ph_volts, atol=1e-6)

    def test_convert_internal_seafet_ph(self):
        internal_ph = dc.convert_internal_seafet_ph(
            ph_counts=self.internal_ph_counts,
            temperature=self.ph_temperature,
            coefs=ec.ph_seafet_internal_coeffs,
        )
        assert np.allclose(internal_ph, self.expected_internal_ph, atol=1e-6)

    def test_convert_external_seafet_ph(self):
        external_ph = dc.convert_external_seafet_ph(
            ph_counts=self.external_ph_counts,
            temperature=self.ph_temperature,
            salinity=35,
            pressure=0,
            coefs=ec.ph_seafet_external_coeffs,
        )
        assert np.allclose(external_ph, self.expected_external_ph, atol=1e-6)



"""A collection of classes and functions related to the processing of instrument data.

Classes:

    InstrumentType (Enum)
    HexDataTypes (Enum)
    Sensors (Enum)
    MeasurementSeries
    InstrumentData

Functions:

    cnv_to_instrument_data (Path) -> InstrumentData
    fix_exponents (List[str]) -> List[str]
    read_hex_file (str, InstrumentType, List[Sensors], bool) -> pd.DataFrame
    preallocate_dataframe (InstrumentType, str, List[Sensors], bool, int) -> pd.DataFrame
    read_hex (InstrumentType, str, List[Sensors], bool) -> dict
    read_SBE19plus_format_0 (str, List[Sensors], bool) -> dict
    read_SBE37SM_format_0 (str, List[Sensors]) -> dict

"""

# Native imports
from enum import Enum
from dataclasses import dataclass
from datetime import date, datetime
from logging import getLogger
from typing import List, Dict
from pathlib import Path

# Third-party imports
import numpy as np
import pandas as pd

# Sea-Bird imports

# Internal imports

logger = getLogger(__name__)

COUNTS_TO_VOLTS = 13107
SECONDS_BETWEEN_EPOCH_AND_2000 = 946684800


class InstrumentType(Enum):
    SBE37SM = "37-SM"
    SBE37SMP = "37-SMP"
    SBE37SMPODO = "37-SMP-ODO"
    SBE37IM = "37-IM"
    SBE37IMP = "37-IMP"
    SBE37IMPODO = "37-IMP-ODO"
    SBE19Plus = "19plus"


class HexDataTypes(Enum):
    temperature = "temperature"
    conductivity = "conductivity"
    pressure = "pressure"
    temperatureCompensation = "temperature compensation"
    ExtVolt0 = "volt 0"
    ExtVolt1 = "volt 1"
    ExtVolt2 = "volt 2"
    ExtVolt3 = "volt 3"
    ExtVolt4 = "volt 4"
    ExtVolt5 = "volt 5"
    SBE38temperature = "SBE38 temperature"
    wetlabs0 = "wetlabs - channel 0"
    wetlabs1 = "wetlabs - channel 1"
    wetlabs2 = "wetlabs - channel 2"
    GTDpressure = "GTD pressure"
    GTDtemperature = "GTD temperature"
    GTDpressure2 = "GTD pressure - sensor 2"
    GTDtemperature2 = "GTD temperature - sensor 2"
    optodeOxygen = "optode oxygen"
    SBE63phase = "SBE63 oxygen phase"
    SBE63temperature = "SBE63 oxygen temperature"
    dateTime = "date time"


# export type HexDataTypeStrings = keyof typeof HexDataTypes;

HEX_LENGTH = {
    "temperature": 6,
    "conductivity": 6,
    "pressure": 6,
    "temperatureCompensation": 4,
    "voltage": 4,
    "SBE38temperature": 6,
    "wetlabsSingleSensor": 4,  # There are three of these for each WL Sensor
    "GTDpressure": 8,
    "GTDtemperature": 6,
    "optodeOxygen": 6,
    "SBE63phase": 6,
    "SBE63temperature": 6,
    "SeaFETVint": 6,
    "SeaFETVext": 6,
    "SeaOWLChannel": 4,  # There are three of these for each SeaOWL
    "time": 8,
}


class Sensors(Enum):
    Temperature = "Temperature"
    Conductivity = "Conductivity"
    Pressure = "Pressure"
    ExtVolt0 = "ExtVolt0"
    ExtVolt1 = "ExtVolt1"
    ExtVolt2 = "ExtVolt2"
    ExtVolt3 = "ExtVolt3"
    ExtVolt4 = "ExtVolt4"
    ExtVolt5 = "ExtVolt5"
    WETLABS = "WETLABS"
    GTD = "GTD"
    DualGTD = "DualGTD"
    OPTODE = "OPTODE"
    SBE63 = "SBE63"
    SBE38 = "SBE38"
    SeaFET = "SeaFET"


@dataclass
class MeasurementSeries:  # TODO make this inherit from a pandas DataFrame
    """Container for measurement data."""

    label: str
    description: str
    units: str
    start_time: date
    values: np.ndarray


@dataclass
class InstrumentData:
    measurements: Dict  # TODO make this inherit from a pandas DataFrame
    interval_s: float
    latitude: float
    start_time: date
    sample_count: int


def cnv_to_instrument_data(filepath: Path) -> InstrumentData:
    """Import the data from a .cnv file and put it into an InstrumentData object.

    Args:
        filepath (Path): the path to the .cnv file to be imported

    Returns:
        InstrumentData: the imported data from the .cnv file

    # TODO: make this an InstrumentData class method
    """

    data = InstrumentData(
        measurements={},
        interval_s=None,
        latitude=0.0,
        start_time=None,
        sample_count=None,
    )

    n = 0

    logger.info("Unpacking instrument data from file: %s", filepath)

    with open(filepath, "r") as cnv:
        for line in cnv:
            if line.startswith("*") or line.startswith("#"):
                if line.startswith("# nvalues = "):
                    data.sample_count = int(line[line.find("= ") + 2 : line.find("\n")])
                elif line.startswith("# name "):
                    label = line[line.find("= ") + 2 : line.find(":")]
                    left_bracket = line.find(" [")
                    if left_bracket > 0:
                        description = line[line.find(": ") + 2 : left_bracket]
                        units = line[line.find("[") + 1 : line.find("]")]
                    else:
                        description = line[line.find(": ") + 2 : line.find("\n")].strip()
                        units = ""

                    data.measurements[label] = MeasurementSeries(
                        label=label,
                        description=description,
                        units=units,
                        start_time=None,
                        values=np.zeros(data.sample_count),
                    )

                elif line.startswith("# interval = "):
                    interval = float(
                        line[line.find(": ") + 2 : line.find("\n")]
                    )  # TODO: fix for minutes, hours, etc
                    data.interval_s = interval

                elif line.startswith("# start_time = "):
                    end = min(line.find("\n"), line.find(" ["))
                    date_string = line[line.find("= ") + 2 : end]
                    start_time = datetime.strptime(date_string, "%b %d %Y %H:%M:%S")
                    data.start_time = start_time
                    for measurement in data.measurements.values():
                        measurement.start_time = start_time

                elif line.startswith("** Latitude: "):
                    latitude_parts = line[line.find(": ") + 2 :].split()
                    data.latitude = float(latitude_parts[0]) + float(latitude_parts[1]) / 60.0
                    # TODO: add higher priority latitude to individual measurement series where necessary

            else:
                values = fix_exponents(" -".join(line.split("-")).split())

                if len(values) > 0:
                    for value, measurement in zip(values, data.measurements.values()):
                        measurement.values[n] = float(value)
                    n += 1
    return data


def fix_exponents(values: List[str]) -> List[str]:
    """Fixes flag values and other numbers with negative exponents.

    The fixes are performed by merging numbers that end in 'e' with the
        following number in the list (the exponent), then deleting the exponent

    Args:
        values (List[str]): List of strings representing numbers

    Returns:
        List[str]: List of strings where eponents have been fixed
    """

    del_indices = [n + 1 for n, value in enumerate(values) if value.endswith("e")]
    for n in del_indices:
        values[n - 1] = f"{values[n-1]}{values[n]}"
    return np.delete(values, del_indices)


def read_hex_file(
    filepath: str,
    instrument_type: InstrumentType,
    enabled_sensors: List[Sensors],
    moored_mode=False,
) -> pd.DataFrame:
    data_count = 0
    is_data = False

    # iterating over file twice in order to preallocate arrays
    file = open(filepath, "r")
    for line in file:
        if is_data and not (line == "" or line.startswith("\n") or line.startswith("\r")):
            data_count += 1
        if line.startswith("*END*"):
            is_data = True

    data_length = data_count
    file.seek(0)
    data_count = 0
    is_data = False

    for line in file:
        if is_data and not (line == "" or line.startswith("\n") or line.startswith("\r")):
            if data_count == 0:
                data = preallocate_dataframe(
                    instrument_type, line, enabled_sensors, moored_mode, data_length
                )
            hex_data = read_hex(instrument_type, line, enabled_sensors, moored_mode)
            data.iloc[data_count] = pd.DataFrame(hex_data, index=[data_count])
            data_count += 1
        if line.startswith("*END*"):
            is_data = True

    file.close()

    return data


def preallocate_dataframe(
    instrument_type: InstrumentType,
    line: str,
    enabled_sensors: List[Sensors],
    moored_mode: bool,
    data_length: int,
) -> pd.DataFrame:
    sensors = {}
    hex_data = read_hex(instrument_type, line, enabled_sensors, moored_mode)
    hex_keys = pd.DataFrame(hex_data, index=[0]).columns
    for key in hex_keys:
        sensors[key] = np.zeros(data_length)

    return pd.DataFrame(sensors)


def read_hex(
    instrument_type: InstrumentType,
    hex: str,
    enabled_sensors: List[Sensors],
    moored_mode=False,
) -> dict:
    """Converts an instrument data hex string into engineering units.

    Args:
        instrument_type (InstrumentType): determines how units are converted
        hex (str): one line from a hex data file
        enabled_sensors (List[Sensors]): mooredMode parses time for 19plus in moored mode if true
        moored_mode (bool, optional): array of Sensors that are enabled. For 37 this is always
            temperature, conductivity, pressure. Defaults to False.

    Returns:
        dict: the sensor values in engineering units that were extracted from the input hex string
    """

    if instrument_type == InstrumentType.SBE19Plus:
        return read_SBE19plus_format_0(hex, enabled_sensors, moored_mode)

    if instrument_type == InstrumentType.SBE37SM:
        return read_SBE37SM_format_0(hex, enabled_sensors)


def read_SBE19plus_format_0(hex: str, enabled_sensors: List[Sensors], moored_mode=False) -> dict:
    """Converts a 19plus V2 data hex string into engineering units.

    Args:
        hex (str): one line from a hex data file
        enabled_sensors (List[Sensors]): mooredMode parses time for 19plus in moored mode if true
        moored_mode (bool, optional): array of Sensors that are enabled. For 37 this is always temperature, conductivity, pressure. Defaults to False.

    Returns:
        dict: the 19plus V2 sensor values in engineering units that were extracted from the input hex string

    Raises:
        RuntimeWarning: if the hex string length does not match the expected length
    """

    results = {}
    n = 0
    for sensor in Sensors:
        if sensor in enabled_sensors:
            if sensor == Sensors.Temperature:
                results[HexDataTypes.temperature.value] = int(
                    hex[n : HEX_LENGTH["temperature"]], 16
                )
                n += HEX_LENGTH["temperature"]

            if sensor == Sensors.Conductivity:
                results[HexDataTypes.conductivity.value] = (
                    int(hex[n : n + HEX_LENGTH["conductivity"]], 16) / 256
                )
                n += HEX_LENGTH["conductivity"]

            if sensor == Sensors.Pressure:  # TODO: add conversion for quartz pressure sensors
                results[HexDataTypes.pressure.value] = int(hex[n : n + HEX_LENGTH["pressure"]], 16)
                n += HEX_LENGTH["pressure"]
                result = (
                    int(hex[n : n + HEX_LENGTH["temperatureCompensation"]], 16) / COUNTS_TO_VOLTS
                )
                results[HexDataTypes.temperatureCompensation.value] = result
                n += HEX_LENGTH["temperatureCompensation"]

            if sensor in [
                Sensors.ExtVolt0,
                Sensors.ExtVolt1,
                Sensors.ExtVolt2,
                Sensors.ExtVolt3,
                Sensors.ExtVolt4,
                Sensors.ExtVolt5,
            ]:
                result = int(hex[n : n + HEX_LENGTH["voltage"]], 16) / COUNTS_TO_VOLTS
                results[HexDataTypes[Sensors[sensor.value].value].value] = result
                n += HEX_LENGTH["voltage"]

            if sensor == Sensors.SBE38:
                results[HexDataTypes.SBE38temperature.value] = (
                    int(hex[n : n + HEX_LENGTH["SBE38temperature"]], 16) / 100000 - 10
                )
                n += HEX_LENGTH["SBE38temperature"]

            if sensor == Sensors.WETLABS:
                results[HexDataTypes.wetlabs0.value] = int(
                    hex[n : n + HEX_LENGTH["wetlabsSingleSensor"]], 16
                )
                n += HEX_LENGTH["wetlabsSingleSensor"]

                results[HexDataTypes.wetlabs1.value] = int(
                    hex[n : n + HEX_LENGTH["wetlabsSingleSensor"]], 16
                )
                n += HEX_LENGTH["wetlabsSingleSensor"]

                results[HexDataTypes.wetlabs2.value] = int(
                    hex[n : n + HEX_LENGTH["wetlabsSingleSensor"]], 16
                )
                n += HEX_LENGTH["wetlabsSingleSensor"]

            if sensor == Sensors.GTD:
                results[HexDataTypes.GTDpressure.value] = (
                    int(hex[n : n + HEX_LENGTH["GTDpressure"]], 16) / 10000
                )
                n += HEX_LENGTH["GTDpressure"]
                results[HexDataTypes.GTDtemperature.value] = (
                    int(hex[n : n + HEX_LENGTH["GTDtemperature"]], 16) / 10000 - 10
                )
                n += HEX_LENGTH["GTDtemperature"]

            if sensor == Sensors.DualGTD:
                results[HexDataTypes.GTDpressure.value] = (
                    int(hex[n : n + HEX_LENGTH["GTDpressure"]], 16) / 10000
                )
                n += HEX_LENGTH["GTDpressure"]
                results[HexDataTypes.GTDtemperature.value] = (
                    int(hex[n : n + HEX_LENGTH["GTDtemperature"]], 16) / 10000 - 10
                )
                n += HEX_LENGTH["GTDtemperature"]
                results[HexDataTypes.GTDpressure2.value] = (
                    int(hex[n : n + HEX_LENGTH["GTDpressure"]], 16) / 10000
                )
                n += HEX_LENGTH["GTDpressure"]
                results[HexDataTypes.GTDtemperature2.value] = (
                    int(hex[n : n + HEX_LENGTH["GTDtemperature"]], 16) / 10000 - 10
                )
                n += HEX_LENGTH["GTDtemperature"]

            if sensor == Sensors.OPTODE:
                results[HexDataTypes.optodeOxygen.value] = (
                    int(hex[n : n + HEX_LENGTH["optodeOxygen"]], 16) / 10000 - 10
                )
                n += HEX_LENGTH["optodeOxygen"]

            if sensor == Sensors.SBE63:
                results[HexDataTypes.SBE63phase.value] = (
                    int(hex[n : n + HEX_LENGTH["SBE63phase"]], 16) / 100000 - 10
                )
                n += HEX_LENGTH["SBE63phase"]
                results[HexDataTypes.SBE63temperature.value] = (
                    int(hex[n : n + HEX_LENGTH["SBE63temperature"]], 16) / 1000000 - 1
                )
                n += HEX_LENGTH["SBE63temperature"]

    if moored_mode:
        seconds_since_2000 = int(hex[n : n + HEX_LENGTH["time"]], 16)
        results[HexDataTypes.dateTime.value] = datetime.fromtimestamp(
            seconds_since_2000 + SECONDS_BETWEEN_EPOCH_AND_2000
        )
        n += HEX_LENGTH["time"]

    # Validate hex length. Ensure length matches what is expected based on enabeld sensors and moored mode.
    if n != len(hex.split("\n")[0]):
        raise RuntimeWarning(
            "Hex string length does not match expectation based on enabled sensors and moored mode"
        )

    return results


def read_SBE37SM_format_0(hex: str, enabled_sensors: List[Sensors]) -> dict:
    """Converts a 37 family data hex string into engineering units.

    Args:
        hex (str): one line from a hex data file
        enabled_sensors (List[Sensors]): mooredMode parses time for 19plus in moored mode if true

    Returns:
        dict: the 37 family sensor values in engineering units that were extracted from the input hex string
    """
    results = {}
    n = 0
    results[HexDataTypes.temperature.value] = int(hex[n : HEX_LENGTH["temperature"]], 16)
    n += HEX_LENGTH["temperature"]

    results[HexDataTypes.conductivity.value] = (
        int(hex[n : n + HEX_LENGTH["conductivity"]], 16) / 256
    )
    n += HEX_LENGTH["conductivity"]

    if Sensors.SBE63 in enabled_sensors:
        results[HexDataTypes.SBE63phase.value] = (
            int(hex[n : n + HEX_LENGTH["SBE63phase"]], 16) / 100000 - 10
        )
        n += HEX_LENGTH["SBE63phase"]
        results[HexDataTypes.SBE63temperature.value] = (
            int(hex[n : n + HEX_LENGTH["SBE63temperature"]], 16) / 1000000 - 1
        )
        n += HEX_LENGTH["SBE63temperature"]

    if Sensors.Pressure in enabled_sensors:
        results[HexDataTypes.pressure.value] = int(hex[n : n + HEX_LENGTH["pressure"]], 16)
        n += HEX_LENGTH["pressure"]
        result = int(hex[n : n + HEX_LENGTH["temperatureCompensation"]], 16)
        results[HexDataTypes.temperatureCompensation.value] = result
        n += HEX_LENGTH["temperatureCompensation"]

    seconds_since_2000 = int(hex[n : n + HEX_LENGTH["time"]], 16)
    results[HexDataTypes.dateTime.value] = datetime.fromtimestamp(
        seconds_since_2000 + SECONDS_BETWEEN_EPOCH_AND_2000
    )
    n += HEX_LENGTH["time"]

    return results



#!/usr/bin/python3
# -*- coding: utf-8 -*-

""" A collection of utility functions related to the processing of SBS instrument data.

Functions:

    close_enough (np.ndarray, np.ndarray, int, float) -> bool
    plot (np.ndarray)
    percent_match (np.ndarray, np.ndarray) -> str

"""

# Native imports

# Third-party imports
import matplotlib.pyplot as plt
import numpy as np

# Sea-Bird imports

# Internal imports


def close_enough(
    test_values: np.ndarray,
    expected_values: np.ndarray,
    rounding_order: int,
    absolute_tolerance: float,
) -> bool:
    """Compares ndarrays, accounting for rounding errors introduced by SeaSoft.

    Frequently a float will be accurate to the 14th decimal, for example, but the
    15th decimal may toggle it over or under 5, causing it to round differently.
    This function adds or subtracts half of the least significant digit, then
    compares values to a given tolerance.

    Args:
        test_values (np.ndarray): values derived by this library
        expected_values (np.ndarray): values derived by SeaSoft
        rounding_order (int): The order that SeaSoft values were rounded to
        absolute_tolerance (float): values must be at least this close after
        adding or subtracting the rounding error

    Returns:
        bool: pass or fail aggregate
    """

    results = np.full(len(test_values), False)
    for n in range(len(test_values)):
        results[n] = (
            np.round(test_values[n], rounding_order) == expected_values[n]
            or np.isclose(
                test_values[n],
                expected_values[n] - 0.5 * 10**-rounding_order,
                rtol=0,
                atol=absolute_tolerance,
            )
            or np.isclose(
                test_values[n],
                expected_values[n] + 0.5 * 10**-rounding_order,
                rtol=0,
                atol=absolute_tolerance,
            )
        )
    return np.all(results)


def plot(**kwargs: np.ndarray):
    """Plots a dictionary of ndarrays

    Args:
        **kwargs (np.ndarray): the dictionary to plot
    """

    fig, ax = plt.subplots(figsize=(20, 10))
    for key in kwargs.keys():
        x = range(len(kwargs[key]))
        ax.plot(x, kwargs[key], label=key)
        ax.legend()
    plt.show()


def percent_match(x1: np.ndarray, x2: np.ndarray) -> str:
    """Calculates the extent that two arrays match.

    Args:
        x1 (np.ndarray): first array to be compared
        x2 (np.ndarray): second array to be compared

    Returns:
        str: a message with the percentage of matching elements
    """

    return f"{100 - (x1 != x2).sum()*100 / len(x1):0.2f}% match"



"""A collection of functions for processing converted SBS instrument data.

Classes:

    MinVelocityType (Enum)
    WindowFilterType (Enum)

Functions:

    butterworth_filter (np.ndarray, float, float) -> np.ndarray
    low_pass_filter (np.ndarray, float, float) -> np.ndarray
    align_ctd (np.ndarray, float, float, float) -> np.ndarray
    cell_thermal_mass (np.ndarray, np.ndarray, float, float, float) -> np.ndarray
    loop_edit_pressure (np.ndarray, float, np.ndarray, float, MinVelocityType, float, float, float, bool, float, float, bool, bool,float) -> np.ndarray
    loop_edit_depth (np.ndarray, np.ndarray, float, MinVelocityType, float, float, float, bool, float, float, bool, bool,float) -> np.ndarray
    find_depth_peaks (np.ndarray, np.ndarray, bool, float, float, float) -> tuple[int, int]
    min_velocity_mask (np.ndarray, float, float, int, int, bool) -> np.ndarray
    mean_speed_percent_mask (np.ndarray, float, float, float, int, int, bool, int) -> np.ndarray
    flag_by_minima_maxima (np.ndarray, np.ndarray, int, int, float)
    bin_average (pd.DataFrame, str, float, bool, int, int, bool, bool)
    wild_edit (np.ndarray, np.ndarray, float, float, int, float, bool,float) -> np.ndarray
    flag_data (np.ndarray, np.ndarray, float, float, float, bool,float) -> np.ndarray
    window_filter (np.ndarray, np.ndarray, WindowFilterType, int, float, int, float, bool, float) -> np.ndarray
    bouyancy_frequency (np.ndarray, np.ndarray, np.ndarray, float)
    calc_N2_buoyancy_retro (np.ndarray, np.ndarray, np.ndarray, float)
    buoyancy (np.ndarray, np.ndarray, np.ndarray, np.ndarray, np.ndarray, float, bool,float)
    Density (np.ndarray, np.ndarray, np.ndarray) -> np.ndarray
    PoTemp (np.ndarray, np.ndarray, np.ndarray, np.ndarray) -> np.ndarray
    ATG (np.ndarray, np.ndarray, np.ndarray) -> np.ndarray

"""

# Native imports
from enum import Enum
import math

# Third-party imports
import gsw
import numpy as np
import pandas as pd
from scipy import signal, stats

# Sea-Bird imports

# Internal imports
from .conversion import depth_from_pressure
from . import eos80_processing as eos80


class MinVelocityType(Enum):
    FIXED = 0
    PERCENT = 1


class WindowFilterType(Enum):
    BOXCAR = "boxcar"
    COSINE = "cosine"
    TRIANGLE = "triangle"
    GAUSSIAN = "gaussian"
    MEDIAN = "median"


def butterworth_filter(x: np.ndarray, time_constant: float, sample_interval: float) -> np.ndarray:
    """Applies a butterworth filter to a dataset.

    Args:
        x (np.ndarray): A numpy array of floats
        time_constant (float): 1 / (2 * pi * cutoff_frequency)
        sample_interval (float): 1 / sampling_frequency

    Returns:
        np.ndarray: the filtered data
    """

    cutoff_frequency = 1 / (2 * np.pi * time_constant)
    sampling_frequency = 1 / sample_interval
    b, a = signal.butter(1, cutoff_frequency, fs=sampling_frequency)
    y = signal.filtfilt(b, a, x)
    return y


def low_pass_filter(x: np.ndarray, time_constant: float, sample_interval: float) -> np.ndarray:
    """Applies a low pass filter as defined in the SeaSoft manual v7.26.8 page 101.

    Args:
        x (np.ndarray): A numpy array of floats
        time_constant (float): 1 / (2 * pi * cutoff_frequency)
        sample_interval (float): 1 / sampling_frequency

    Returns:
        np.ndarray: the filtered data
    """

    a = 1 / (1 + 2 * time_constant / sample_interval)
    b = a * (1 - 2 * time_constant / sample_interval)

    x2 = x.copy()
    for n in range(1, len(x)):
        x2[n] = a * x[n] + a * x[n - 1] - b * x2[n - 1]

    y = x2.copy()
    for n in range(len(x) - 2, -1, -1):
        y[n] = a * x2[n] + a * x2[n + 1] - b * y[n + 1]

    return y


def align_ctd(
    x: np.ndarray, offset: float, sample_interval: float, flag_value=-9.99e-29
) -> np.ndarray:
    """Takes an ndarray object of data for a single variable and applies a time offset to the series.

    Performs interpolation when offset is not a factor of sampleinterval

    Args:
        x (np.ndarray): array of values to apply shift to
        offset (float): offset value to shift by (s)
        sample_interval: time between samples (s)

    Returns:
        np.ndarray: the aligned data

    TODO: See JIRA issue NSI-1658
    """

    sample_times = np.arange(0, len(x) * sample_interval, sample_interval)
    interp_res = np.interp(sample_times + offset, sample_times, x, flag_value, flag_value)
    return interp_res


def cell_thermal_mass(
    temperature_C: np.ndarray,
    conductivity_Sm: np.ndarray,
    amplitude: float,
    time_constant: float,
    sample_interval: float,
) -> np.ndarray:
    """Removes conductivity cell thermal mass effects from measured conductivity.

    Cell Thermal Mass uses a recursive filter to remove conductivity cell thermal
        mass effects from the measured conductivity [From the SeaSoft manual, page 92]

    Args:
        temperature_C (np.ndarray): temperature in degrees C
        conductivity_Sm (np.ndarray): conductivity in S/m
        amplitude (float): thermal anomaly amplitude (alpha)
        time_constant (float): thermal anomoly time constant (1/beta)
        sample_interval (float): time between samples

    Returns:
        np.ndarray: the corrected conductivity in S/m
    """

    a = 2 * amplitude / (sample_interval / time_constant + 2)
    b = 1 - (2 * a / amplitude)
    ctm = np.zeros(len(temperature_C))  # cell thermal mass
    corrected_conductivity = conductivity_Sm.copy()

    for n in range(1, len(ctm)):
        dc_dT = 0.1 * (1.0 + 0.006 * (temperature_C[n] - 20.0))
        dT = temperature_C[n] - temperature_C[n - 1]
        ctm[n] = -1.0 * b * ctm[n - 1] + a * dc_dT * dT
        corrected_conductivity[n] += ctm[n]

    return corrected_conductivity


def loop_edit_pressure(
    pressure: np.ndarray,
    latitude: float,
    flag: np.ndarray,
    sample_interval: float,
    min_velocity_type: MinVelocityType,
    min_velocity: float,
    window_size: float,
    mean_speed_percent: float,
    remove_surface_soak: bool,
    min_soak_depth: float,
    max_soak_depth: float,
    use_deck_pressure_offset: bool,
    exclude_flags: bool,
    flag_value=-9.99e-29,
) -> np.ndarray:
    """Variation of loop_edit_depth that derives depth from pressure and latitude.

    Args:
        pressure (np.ndarray): A pressure array in dbar
        latitude (float): A single latitude value where the cast occurred

        See loop_edit_depth for remaining args

    Returns:
        np.ndarray: the input data with updated flags
    """

    depth = depth_from_pressure(pressure, latitude)
    return loop_edit_depth(
        depth,
        flag,
        sample_interval,
        min_velocity_type,
        min_velocity,
        window_size,
        mean_speed_percent,
        remove_surface_soak,
        min_soak_depth,
        max_soak_depth,
        use_deck_pressure_offset,
        exclude_flags,
        flag_value,
    )


def loop_edit_depth(
    depth: np.ndarray,
    flag: np.ndarray,
    sample_interval: float,
    min_velocity_type: MinVelocityType,
    min_velocity: float,
    window_size: float,
    mean_speed_percent: float,
    remove_surface_soak: bool,
    min_soak_depth: float,
    max_soak_depth: float,
    use_deck_pressure_offset: bool,
    exclude_flags: bool,
    flag_value=-9.99e-29,
) -> np.ndarray:
    """Marks scans determined to be part of a pressure loop as bad.

    Loop Edit marks scans bad by setting the flag value associated with the scan to
    badflag in data that has pressure slowdowns or reversals (typically
    caused by ship heave).

    Args:
        depth (np.ndarray): Salt water depth as an array of positive numbers.
            Colloquially used interchangeably with pressure in dbar
        flag (np.ndarray): Array of flag values. The flag for a good value is 0.0.
            The flag for a detected loop value defaults to -9.99e-29
        sample_interval (float): Time interval between samples in seconds
        min_velocity_type (MinVelocityType): Sets whether flags are based on min
            velocity or a percentage of mean speed
        min_velocity (float): The minimum velocity for data to be considered good
        window_size (float): Time interval to include in mean speed calculation
        mean_speed_percent (float): Percentage of mean speed for data to be considered good
        remove_surface_soak (bool): If true, data that occur before the minimum
            soak depth is reached are marked as bad
        min_soak_depth (float): The depth that must be reached before data can be considered good
        max_soak_depth (float): The maximum depth that can be considered the start of a downcast
        use_deck_pressure_offset (bool): If true, min and max soak depths are offset
            by the first value in the depth array
        exclude_flags (bool): If true, existing bad flags are preserved
        flag_value (_type_, optional): Passing is 0.0, failing defaults to -9.99e-29.

    Returns:
        np.ndarray: the input data with updated flags
    """

    if not exclude_flags:
        flag[:] = 0.0

    if use_deck_pressure_offset:
        min_soak_depth -= depth[0]
        max_soak_depth -= depth[0]

    (min_depth_n, max_depth_n) = find_depth_peaks(
        depth, flag, remove_surface_soak, flag_value, min_soak_depth, max_soak_depth
    )

    if min_velocity_type == MinVelocityType.FIXED:
        downcast_mask = min_velocity_mask(
            depth, sample_interval, min_velocity, min_depth_n, max_depth_n + 1, False
        )
        upcast_mask = min_velocity_mask(
            depth, sample_interval, min_velocity, max_depth_n, len(depth), True
        )

    elif min_velocity_type == MinVelocityType.PERCENT:
        diff_length = int(window_size / sample_interval)
        downcast_mask = mean_speed_percent_mask(
            depth,
            sample_interval,
            min_velocity,
            mean_speed_percent,
            min_depth_n,
            max_depth_n,
            False,
            diff_length,
        )
        upcast_mask = mean_speed_percent_mask(
            depth,
            sample_interval,
            min_velocity,
            mean_speed_percent,
            max_depth_n,
            len(depth),
            True,
            diff_length,
        )

    flag[~downcast_mask & ~upcast_mask] = flag_value

    flag_by_minima_maxima(depth, flag, min_depth_n, max_depth_n, flag_value)

    cast = np.array([0 for _ in range(len(flag))])
    cast[downcast_mask] = -1
    cast[upcast_mask] = 1
    return cast  # TODO: refactor to handle cast and flag in seperate functions


def find_depth_peaks(
    depth: np.ndarray,
    flag: np.ndarray,
    remove_surface_soak: bool,
    flag_value: float,
    min_soak_depth: float,
    max_soak_depth: float,
) -> tuple[int, int]:
    """Finds the global depth minima and maxima.

    This determinines the earliest points where the downcast and upcast can begin

    Args:
        depth (np.ndarray): depth data
        flag (np.ndarray): flag data
        remove_surface_soak (bool): If true, scans before a minimum depth are mrked as bad
        flag_value (float): the flag value (typically -9.99e-29)
        min_soak_depth (float): the minimum depth that must be reached before a series can be considered a downcast
        max_soak_depth (float): maximumm depth that can be considered the start of a downcast

    Returns:
        tuple[int, int]: minimum and maximum index corresponding to minimum and maximum depth
    """

    # first index where min_soak_depth < depth < max_soak_depth
    if remove_surface_soak:
        min_soak_depth_n = min(
            n
            for n, d in enumerate(depth)
            if flag[n] != flag_value and min_soak_depth < d and d < max_soak_depth
        )
    else:
        min_soak_depth_n = 0

    max_soak_depth_n = min(
        n for n, d in enumerate(depth) if flag[n] != flag_value and d > max_soak_depth
    )

    # beginning of possible downcast domain
    min_depth_n = min(
        (d, n)
        for n, d in enumerate(depth)
        if flag[n] != flag_value and min_soak_depth_n <= n and n < max_soak_depth_n
    )[1]

    # beginning of possible upcast domain
    max_depth_n = (
        -1
        if min_depth_n == len(depth) - 1
        else [
            # TODO Refactor. Converting to a dataframe here is expensive
            # n for n,d in enumerate(depth) if d == max(pd.DataFrame(depth).dropna().values) and n > min_depth_n
            n
            for n, d in enumerate(depth)
            if d == max(depth) and n > min_depth_n
        ][0]
    )

    return (min_depth_n, max_depth_n)


def min_velocity_mask(
    depth: np.ndarray,
    interval: float,
    min_velocity: float,
    domain_start: int,
    domain_end: int,
    is_upcast: bool,
) -> np.ndarray:
    """Creates a mask to assign bad flags where velocity is less that the provided minimum.

    Args:
        depth (np.ndarray): depth data
        interval (float): sampling interval
        min_velocity (float): minimum velocity (only used for samples within first window)
        domain_start (int): earliest possible beginning of downcast or upcast
        domain_end (int): earliest possible end of downcast or upcast
        is_upcast (bool): inverts velocity sign for upcast

    Returns:
        np.ndarray: true/false mask where true means velocity is at least the minimum
    """

    sign = -1 if is_upcast else 1
    mask0 = sign * np.diff(depth, prepend=depth[0]) / interval >= min_velocity
    mask1 = sign * np.diff(depth, append=depth[-1]) / interval >= min_velocity

    mask = mask0 & mask1

    mask[:domain_start] = False
    mask[domain_end:] = False

    return mask


def mean_speed_percent_mask(
    depth: np.ndarray,
    interval: float,
    min_velocity: float,
    mean_speed_percent: float,
    domain_start: int,
    domain_end: int,
    is_upcast: bool,
    diff_length: int,
) -> np.ndarray:
    """Assigns bad flags to scans below a specified velocity.

    This creates a mask to assign bad flags where the velocity is less than a
    given percentage of the mean accross a number of samples according to
    diff_length (determined by the window and sample_rate)

    Args:
        depth (np.ndarray): depth data
        interval (float): sampling interval in seconds
        min_velocity (float): minimum velocity
        mean_speed_percent (float): the minimum percentage of the mean speed that qualifies a sample as good
        domain_start (int): earliest possible beginning of downcast
        domain_end (int): earliest possible beginning of upcast
        is_upcast (bool): inverts velocity sign for upcast
        diff_length (int): averaging window divided by sample interval

    Returns:
        np.ndarray: true/false mask where true means the sample velocity is at
            least the given percentage of the mean
    """

    sign = -1 if is_upcast else 1
    mask0 = sign * np.diff(depth[0:diff_length]) / interval > min_velocity
    mask1 = sign * np.diff(depth[1 : diff_length + 1]) / interval > min_velocity
    first_window_mask = np.concatenate([[False], (mask0 & mask1)])

    mean_speed = sign * (depth[diff_length:] - depth[:-diff_length]) / diff_length / interval
    speed = sign * np.diff(depth[diff_length:], prepend=depth[diff_length - 1]) / interval

    mask = np.concatenate((first_window_mask, (speed > (mean_speed * mean_speed_percent / 100.0))))
    mask[:domain_start] = False
    mask[domain_end:] = False

    return mask


def flag_by_minima_maxima(
    depth: np.ndarray,
    flag: np.ndarray,
    min_depth_n: int,
    max_depth_n: int,
    flag_value: float,
):
    """Flags data that is less than the most recent local minima or maxima.

    This condition occurs following a ship heave, where velocity temporarily inverts

    Args:
        depth (np.ndarray): depth data
        flag (np.ndarray): flag data.  Will be modified by this function.
        min_depth_n (int): index of global minimum depth
        max_depth_n (int): index of global maximum depth
        flag_value (float): value assigned to bad flags
    """

    local_max = -10000.0
    local_min = 10000.0

    # flag values that don't exceed most recent valid local minima/maxima
    for n in range(0, len(depth)):
        if n >= max_depth_n and depth[n] < local_min and flag[n] != flag_value:
            local_min = depth[n]
        elif n >= min_depth_n and depth[n] > local_max and flag[n] != flag_value:
            local_max = depth[n]
        else:
            flag[n] = flag_value


def bin_average(
    dataset: pd.DataFrame,
    bin_variable: str,
    bin_size: float,
    include_scan_count: bool,
    min_scans: int,
    max_scans: int,
    exclude_bad_scans: bool,
    interpolate: bool,
):
    """Averages data into a number of set bins, averaging values within each bin.

    Bins with fewer than min_scans or larger than max_scans get removed.
    Updates the dataset argument to be in bins.

    Args:
        dataset (pd.DataFrame): Data frame containing all data to group into bins. Will be modified by this function.
        bin_variable (string): name of the variable to use bins for. Expected to be one of "pressure", "depth", "scan_number", or "time"
        bin_size (number): size for each bin of data
        include_scan_count (bool): indicates whether to include a column in the updated dataset for the number of scans in each bin
        min_scans (number): indicates the minimum number of scans required in a bin for it to be included in the dataset. Note that bins with 0 scans are always dropped.
        max_scans (number): indicates the maximum number of scans allowed to be in a bin for it to be included in the dataset. Bins with more scans than max_Scans will be dropped.
        exclude_bad_scans (boolean): indicates whether to include bad scans within the bins.
        interpolate (boolean): indicates whether to interpolate the balues in the bins after averaging. Only possible for pressure or depth bins

    TODO: Currently unimplemented: surface bin, begin/end scans to omit, processing upcast vs downcast vs both, time bins, scan number bins
    """

    if exclude_bad_scans and "flag" in dataset.columns:
        # remove all scans marked as bad (i.e. flag greater than 0)
        dataset.drop(dataset[dataset["flag"] > 0].index, inplace=True)
    bin_row = dataset[
        bin_variable
    ].to_numpy()  # pd series containing the variable we want to bin for, converted to ndarray
    max_val = np.amax(bin_row)

    bin_min = bin_size - (bin_size / 2.0)  # min value of first bin

    # create the bins to sort into
    # note that we create an inital "dummy" bin from -bin_min to bin_min to catch extra low values
    # TODO: Above issue can be addressed by surface bin likely
    bin_targets = np.arange(start=bin_min - bin_size, stop=max_val + bin_size, step=bin_size)

    # setup bins to indicate where each index should be sorted into
    bins = np.digitize(x=bin_row, bins=bin_targets, right=True)

    dataset["bin_number"] = bins

    dataset = dataset.groupby("bin_number").mean()

    # get the number of scans in each bin
    nbin_unfiltered = np.bincount(bins)
    nbin = np.delete(nbin_unfiltered, np.where(nbin_unfiltered == 0))
    dataset["nbin"] = nbin

    dataset.drop(dataset[dataset["nbin"] < min_scans].index, inplace=True)
    dataset.drop(dataset[dataset["nbin"] > max_scans].index, inplace=True)

    print(dataset.to_string())

    # TODO: validate that this is running correctly
    if interpolate:
        new_dataset = dataset.copy()
        prev_row = None
        first_row = None
        first_row_index = None
        second_row = None
        for index, row in dataset.iterrows():
            if prev_row is None:
                # we'll come back to the first row at the end
                first_row = row
                first_row_index = index
            else:
                prev_presure = prev_row[
                    bin_variable
                ]  # use bin_variable since this could be pressure or depth
                curr_pressure = row[
                    bin_variable
                ]  # use bin_variable since this could be pressure or depth
                center_pressure = index * bin_size

                for col_name, col_data in dataset.items():
                    if col_name == "nbin" or col_name == "flag" or col_name == "bin_number":
                        continue
                    prev_val = prev_row[col_name]
                    curr_val = row[col_name]

                    # formula from the seasoft data processing manual, page 89, version 7.26.8-3
                    interpolated_val = (
                        (curr_val - prev_val)
                        * (center_pressure - prev_presure)
                        / (curr_pressure - prev_presure)
                    ) + prev_val
                    # print('interpolated to ' + str(interpolated_val) + ' for ' + str(col_name) + ', ' + str(index))
                    new_dataset.loc[index, col_name] = interpolated_val
            prev_row = row
            if index == 2:
                # save this so we can reference it at the end for interpolating the first row
                second_row = row

        # back to the first row now
        prev_presure = second_row[bin_variable]  # reference second row's value
        curr_pressure = first_row[
            bin_variable
        ]  # use bin_variable since this could be pressure or depth
        center_pressure = first_row_index * bin_size

        for col_name, col_data in dataset.items():
            if col_name == "nbin" or col_name == "flag" or col_name == "bin_number":
                continue
            prev_val = second_row[col_name]  # reference second row's value
            curr_val = first_row[col_name]

            # formula from the seasoft manual, page 89
            interpolated_val = (
                (curr_val - prev_val)
                * (center_pressure - prev_presure)
                / (curr_pressure - prev_presure)
            ) + prev_val
            new_dataset.loc[first_row_index, col_name] = interpolated_val

        dataset = new_dataset
        print(dataset.to_string())
    if not include_scan_count:
        dataset.drop("nbin", axis=1, inplace=True)


# TODO: move to a unit test?
# data = {'pressure': [0, 0.5, 1, 1.2, 1.3, 1.5, 1.9, 2.5, 2.9, 3.5, 4, 5, 6, 7, 8, 9, 11, 13],
#         'depth': [1, 1.5, 2, 22, 9, 16, 3, 4, 7, 8, 9, 8, 7, 1, 21, 19, 18, 6],
#         'temperature': [0, 0.5, 1, 1.2, 1.3, 1.5, 1.9, 2.5, 2.9, 3.5, 4, 5, 6, 7, 8, 9, 11, 13],
#         'flag': [0, 0, 0, 9999, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 9999, 0, 0, 0],}

# df = pd.DataFrame(data)

# print('binning 2 \n' + df.to_string())
# bin_average(df.copy(), 'depth', 2, True,0,10, False, True)

# print('binning 3 \n' + df.to_string())
# bin_average(df.copy(), 'depth', 3, True,0,10, False, True)

# print('binning 5 \n' + df.to_string())
# bin_average(df.copy(), 'depth', 5, True,0,10, True, True)


def wild_edit(
    data: np.ndarray,
    flags: np.ndarray,
    std_pass_1: float,
    std_pass_2: float,
    scans_per_block: int,
    distance_to_mean: float,
    exclude_bad_flags: bool,
    flag_value=-9.99e-29,
) -> np.ndarray:
    """Flags outliers in a dataset.

    Outliers are flagged by iterating over the data in blocks, taking the
    mean and standard deviation, and flagging data outside the combined variance (standard
    deviation of the block multiplied by the standard deviation argument).
    Each block is processed three times: first to remove loop edit flags if applicable,
    second to temporarily remove outliers outside std_pass_1, third to remove outliers
    outside std_pass_2 from the returned data.

    If the final block contains fewer samples than scans_per_block, data is backfilled
    from the previous block before computing the mean and standard deviation.

    This algorithm may introduce nonlinearities in the data because it runs one block
    at a time instead of a rolling window. This is done to maintain parity with SBE Data Processing.

    Args:
        data (np.ndarray): The data to be flagged, such as temperature, pressure, etc.
        flags (np.ndarray): Flag data from loop edit
        std_pass_1 (float): Standard deviation for the first pass
        std_pass_2 (float): Standard deviation for the second pass
        scans_per_block (int): Number of samples to process in each block
        distance_to_mean (float): Minimum threshod to flag data. Values within this range
        to mean are kept even if their standard deviation is outside the limit
        exclude_bad_flags (bool): Excludes all loop edit flags
        flag_value (any, optional): The flag value written in place of data. Defaults to -9.99e-29.

    Returns:
        np.ndarray: The data with flag values in place of outliers
    """

    lower_index = 0
    upper_index = scans_per_block
    flagged_data = data.copy()

    while upper_index <= len(data):
        flagged_data[lower_index:upper_index] = flag_data(
            data[lower_index:upper_index],
            flags,
            std_pass_1,
            std_pass_2,
            distance_to_mean,
            exclude_bad_flags,
            flag_value,
        )
        lower_index += scans_per_block
        upper_index += scans_per_block

    if lower_index < len(data):
        lower_index = len(data) - scans_per_block
        upper_index = len(data)
        flagged_data[len(data) - len(data) % scans_per_block :] = flag_data(
            data[lower_index:upper_index],
            flags,
            std_pass_1,
            std_pass_2,
            distance_to_mean,
            exclude_bad_flags,
            flag_value,
        )[scans_per_block - len(data) % scans_per_block :]

    return flagged_data


def flag_data(
    data: np.ndarray,
    flags: np.ndarray,
    std_pass_1: float,
    std_pass_2: float,
    distance_to_mean: float,
    exclude_bad_flags: bool,
    flag_value=-9.99e-29,
) -> np.ndarray:
    """Helper function for wild_edit() that handles the three main loops.

    Args:
        data (np.ndarray): The data to be flagged, such as temperature, pressure, etc.
        flags (np.ndarray): Flag data from loop edit
        std_pass_1 (float): Standard deviation for the first pass
        std_pass_2 (float): Standard deviation for the second pass
        distance_to_mean (float): Minimum threshod to flag data. Values within this range
        to mean are kept even if their standard deviation is outside the limit
        exclude_bad_flags (bool): Excludes all loop edit flags
        flag_value (_type_, optional): The flag value written in place of data. Defaults to -9.99e-29.

    Returns:
        np.ndarray: The data with flag values in place of outliers
    """

    data_copy = pd.Series(data.copy())
    flagged_data = data.copy()

    for n, value in enumerate(data_copy):
        if exclude_bad_flags and flags[n] == flag_value:
            data_copy[n] = np.nan

    mean = data_copy.mean()
    std = data_copy.std()

    for n, value in enumerate(data_copy):
        if abs(value - mean) >= std * std_pass_1:
            data_copy[n] = np.nan

    mean = data_copy.mean()
    std = data_copy.std()

    for n, value in enumerate(flagged_data):
        if abs(value - mean) > std * std_pass_2 and abs(value - mean) > distance_to_mean:
            flagged_data[n] = flag_value

    return flagged_data


def window_filter(
    data_in: np.ndarray,
    flags: np.ndarray,
    window_type: WindowFilterType,
    window_width: int,
    sample_interval: float,
    half_width=1,
    offset=0.0,
    exclude_flags=False,
    flag_value=-9.99e-29,
) -> np.ndarray:
    """Filters a dataset by convolving it with an array of weights.

    The available window filter types are boxcar, cosine, triangle, gaussian and median.
    Refer to the SeaSoft data processing manual version 7.26.8, page 108

    Args:
        data_in (np.ndarray): data to be filtered
        flags (np.ndarray): flagged data defined by loop edit
        window_type (WindowFilterType): the filter type (boxcar, cosine, triangle, or gaussian)
        window_width (int): width of the window filter (must be odd)
        sample_interval (float, optional): sample interval of the dataset. Defaults to 1.0.
        half_width (int, optional): width of the guassian curve. Defaults to 1.
        offset (float, optional): shifts the center point of the gaussian. Defaults to 0.0.
        exclude_flags (bool, optional): exclude values from the dataset that are
        flagged in flags. Also excludes corresponding weights from normalization Defaults to False.
        flag_value (any, optional): the flag value in flags. Defaults to -9.99e-29.

    Returns:
        np.ndarray: the convolution of data_in and the window filter
    """

    # convert flags to nan for processing
    data = [d if d != flag_value else np.nan for d in data_in]

    if exclude_flags:
        data = [data[n] if f != flag_value else np.nan for n, f in enumerate(flags)]

    # for simplicity later, window indices start negative, end positive, and are centered on 0
    window_start = int(-(window_width - 1) / 2)
    window_end = int((window_width - 1) / 2 + 1)
    window = np.array([])

    # define the window filter
    if window_type == WindowFilterType.BOXCAR:
        window = signal.windows.boxcar(window_width)

    elif window_type == WindowFilterType.COSINE:
        for n in range(window_start, window_end):
            window = np.append(window, np.cos((n * np.pi) / (window_width + 1)))

    elif window_type == WindowFilterType.TRIANGLE:
        window = signal.windows.triang(window_width)

    elif window_type == WindowFilterType.GAUSSIAN:
        phase = offset / sample_interval
        # the manual defines scale with sample rate, but seasoft uses sample interval
        scale = np.log(2) * (2 * sample_interval / half_width) ** 2
        for n in range(window_start, window_end):
            window = np.append(window, np.exp(-((n - phase) ** 2) * scale))

    data_valid = np.nan_to_num(data)
    data_out = data_valid.copy()
    data_start = int((window_width - 1) / 2)
    data_end = len(data) - data_start

    # run a convolution, renormalizing weights as necessary
    for n in range(len(data)):
        if data_start <= n and n < data_end:
            value = 0

            if window_type == WindowFilterType.MEDIAN:
                window = data[n + window_start : n + window_end]
                value = np.nanmedian(window)
            else:
                window_valid = window.copy()

                # exclude weights from normalization if they correspond to nans in the data
                for n2 in range(len(window)):
                    if np.isnan(data[n + data_start - n2]):
                        window_valid[n2] = 0

                window_valid /= window_valid.sum()

                for n2 in range(len(window)):
                    value += window_valid[n2] * data_valid[n + data_start - n2]

            data_out[n] = value

    return data_out


def bouyancy_frequency(
    temp_conservative_subset: np.ndarray,
    salinity_abs_subset: np.ndarray,
    pressure_dbar_subset: np.ndarray,
    gravity: float,
):
    """Calculates an N^2 value (buoyancy frequency) for the given window of temperature, salinity, and pressure, at the given latitude.

    Expect temperature as conservative temperature, salinity as abslute salinity, and pressure as dbar, all of the same length
    Performs the calculation in a modern way using TEOS-10 and specific volume.

    Args:
        temp_conservative_subset (np.ndarray): temperature values for the given window
        salinity_abs_subset (np.ndarray): salinity values for the given window
        pressure_dbar_subset (np.ndarray): pressure values for the given window
        gravity (float): gravity value

    Returns:
        float: A single N^2 [Brunt-Väisälä (buoyancy) frequency]
    """

    db_to_pa = 1e4
    # Wrap these as a length-1 array so that GSW accepts them
    pressure_bar = [np.mean(pressure_dbar_subset)]
    temperature_bar = [np.mean(temp_conservative_subset)]
    salinity_bar = [np.mean(salinity_abs_subset)]

    # Compute average specific volume, temp expansion ceoff, and saline contraction coeff over window
    (v_bar, alpha_bar, beta_bar) = gsw.specvol_alpha_beta(
        salinity_bar, temperature_bar, pressure_bar
    )

    # Estimate vertical gradient of conservative temp
    dCTdp_result = stats.linregress(pressure_dbar_subset, temp_conservative_subset)
    # TODO: error handling with r, p, std_error

    # Estimate vertical gradient of absolute salinity
    dSAdp_result = stats.linregress(pressure_dbar_subset, salinity_abs_subset)
    # TODO: error handling with r, p, std_error

    # Compute N2 combining computed ceofficients and vertical gradients.
    # we index into v_bar, alpha_bar, and beta_bar as they are all arrays of len 1
    n2 = gravity**2 / (v_bar[0] * db_to_pa)
    n2 *= beta_bar[0] * dSAdp_result.slope - alpha_bar[0] * dCTdp_result.slope
    return n2


def buoyancy(
    temperature_c: np.ndarray,
    salinity_prac: np.ndarray,
    pressure_dbar: np.ndarray,
    latitude: np.ndarray,
    longitude: np.ndarray,
    window_size: float,
    use_modern_formula=True,
    flag_value=-9.99e-29,
):
    """Calculates the 4 buoyancy values based off the incoming data.

    Data is expected to have already been binned via Bin_Average using decibar pressure bins.
    All arrays are expected to be the same length, except for latitude and longitude, which can be length 1.
    Optionally can use the former calculation for N^2 from the SBE Data Processing Manual, but defaults to a newer formula using TEOS-10.

    Args:
        temperature_c (np.ndarray): Temperature in ITS-90 degrees C
        salinity_prac (np.ndarray): Practical salinity in PSU
        pressure_dbar (np.ndarray): Pressure in dbar
        latitude (np.ndarray): latitude values. If length 1, gets applied to all values.
        longitude (np.ndarray): longitude values. If length 1, gets applied to all values.
        window_size (float): window size to use.
            If this number is smaller than the binned window size, round up to a minium of 3 scans.
            I.E. uses the center scan and one scan on each side of it at the very least
        use_modern_formula (bool): Whether to use a modern formula for calculating N^2. Defaults to true.
        flag_value (float): Bad Flag value to use for marking bad scans. Defaults to -9.99e-29

    Returns:
        pd.DataFrame: dataframe with 4 columns, one with each calculated variable. The columns are as follows:
    """

    salinity_prac, temperature_c, pressure_dbar, latitude, longitude = np.broadcast_arrays(
        salinity_prac, temperature_c, pressure_dbar, latitude, longitude
    )
    pressure_dbar = pressure_dbar.astype(np.double)

    # Get the original bin size that we're working with, using the second and third bin so we don't have to worry about the surface bin
    original_bin_size = abs(pressure_dbar[2] - pressure_dbar[1])

    # Calculates how many scans to have on either side of our median point, but need at least 1 (for a total of 3 scans)
    scans_per_side = max(math.floor(window_size / original_bin_size / 2), 1)

    salinity_abs = gsw.SA_from_SP(salinity_prac, pressure_dbar, longitude, latitude)
    temperature_conservative = gsw.CT_from_t(salinity_abs, temperature_c, pressure_dbar)

    result = pd.DataFrame()

    # create our result np.ndarrays with the flag value as default
    result["N2"] = np.full(len(temperature_c), flag_value)
    result["N"] = np.full(len(temperature_c), flag_value)
    result["E"] = np.full(len(temperature_c), flag_value)
    result["E10^-8"] = np.full(len(temperature_c), flag_value)

    # start loop at scans_per_side
    for i in range(scans_per_side, len(temperature_conservative) - scans_per_side):
        min_index = i - scans_per_side
        max_index = (
            i + scans_per_side + 1
        )  # add + 1 because slicing does not include the max_index

        pressure_subset = pressure_dbar[min_index:max_index]
        temperature_cons_subset = temperature_conservative[min_index:max_index]
        temperature_ITS_subset = temperature_c[min_index:max_index]
        salinity_subset = salinity_abs[min_index:max_index]

        pressure_bar = [np.mean(pressure_subset)]
        gravity = gsw.grav([latitude[i]], pressure_bar)[0]

        if use_modern_formula:
            salinity_subset = salinity_abs[min_index:max_index]
            n2 = bouyancy_frequency(
                temperature_cons_subset, salinity_subset, pressure_subset, gravity
            )
        else:
            salinity_subset = salinity_prac[min_index:max_index]
            n2 = eos80.bouyancy_frequency(
                temperature_ITS_subset, salinity_subset, pressure_subset, gravity
            )

        result.at[i, "N2"] = n2
        if n2 >= 0:
            result.at[i, "N"] = math.sqrt(n2) * 3600 / (2 * np.pi)
        else:
            result.at[i, "N"] = np.nan
        result.at[i, "E"] = n2 / gravity
        result.at[i, "E10^-8"] = result.at[i, "E"] * 1e8

    return result
